# $Id: TrafD.pm,v 1.2 2004/04/14 09:35:31 stas_degteff Exp $
# 
# Copyright 2003-2004 Dmitry Y. Kazarov. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1.  Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer. 
# 2.  Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution. 
# 
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package TrafD;
use strict;

=head1 NAME

TrafD - trafd savefile interface for Perl

=head1 SYNOPSIS

Reading whole save at once.

  use TrafD;

  my $trafdSave = TrafD->open($fileName) or
    die "Can not open file $fileName: ".$TrafD::errstr;

  while(
      my (
	$NumOfRecords,
	$StartTimeSec,$StartTimeUSec,
	$EndTimeSec,$EndTimeUSec,
	@Records
      ) = $trafdSave->ReadSave
  ) {
      foreach( @Records ) {
	  my (
	      $src_ip, $dst_ip,
	      $proto,
	      $src_port, $dst_port,
	      $bytes, $dataBytes
	  ) = @$_;
      }
  }

  if( $trafdSave->errstr )
    print "error ",$trafdSave->errstr;

  exit 0;

Reading save record by record

  use TrafD;

  my $trafdSave = TrafD->open($fileName) or
    die "Can not open file $fileName: ".$TrafD::errstr;

  LOOP: while(
      my (
	$NumOfRecords,
	$StartTimeSec,$StartTimeUSec,
	$EndTimeSec,$EndTimeUSec
      ) = $trafdSave->ReadHeader
  ) {
      for( my $cntr=0; $cntr<$NumOfRecords; $cntr++ ) {
	  my (
	      $src_ip, $dst_ip,
	      $proto,
	      $src_port, $dst_port,
	      $bytes, $dataBytes
	  ) = $trafdSave->ReadRecord;

	  if( $trafdSave->errstr ) {
	    print "error ",$trafdSave->errstr;
	    break LOOP;
	  }

      }
  }

Other functions

 my $trafd3Save = TrafD3->open($filename);
 my $trafd4Save = TrafD4->open($filename);
 my $stringIPAddr = TrafD::InetN2A($src_ip);

=head1 DESCRIPTION

The TrafD is a L<trafd> save file read module for the Perl programming language
L<trafd> save file consists of one or more saves. Each save has a header and
list of records.

=over

=item C<TrafD-E<gt>open>

opens a trafd save file detecting trafd version. Versions 3 and 4 of save files 
are supported. Mixing of version 3 and 4 saves withing the single file is not allowed.

=item C<TrafD3-E<gt>open>

opens a trafd version 3 save file (old format of file) without check of version of saves.

=item C<TrafD4-E<gt>open>

opens a trafd version 4 (bpft) save file without check of version of saves.

=item C<ReadHeader>

  my ($NumberOfRecords,$StartTime,$StartTimeUSec,$EndTime,$EndTimeUSec) =
    $trafdSave->ReadHeader;

Reads header at current position of file. The file position pointer should be
located on the save header (reading all records of the save moves the file
position pointer to the header of next save).

Returns array of 5 integers: number or records in this save, start time
of collecting traffic - seconds and microseconds, end time of collecting
traffic - seconds and microseconds.

=item C<ReadRecord>

  my(
    $src_ip, $dst_ip,
    $protocol,
    $src_port, $dst_port,
    $bytes, $dataBytes
  ) = $trafdSave->ReadRecord;

Reads single record of the collected data. 

Returns 7 integers: source and destination IP addresses in single number form
(i.e. '172.16.0.1' will be represented as 2886729729), IP protocol number
(refer to /etc/protocols), source and destination port number (refer to
/etc/services), number of bytes passed (including protocol and IP headers),
number of data bytes passed (payload of UDP/TCP)

=item C<ReadSave>

      my (
	$NumOfRecords,
	$StartTimeSec,$StartTimeUSec,
	$EndTimeSec,$EndTimeUSec,
	@Records
      ) = $trafdSave->ReadSave

Reads the save header and all save records. Returns all header fields and array
of records arrays.

I.e. ReadSave is similar to:

    my ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec ) =
		$this->ReadHeader;
    my @Records;
    for( my $cntr=0;$cntr<$NoRec;$cntr++ ) {
	push @Records, [ @record ];
    }
    return ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec,@Records );

=item C<InetN2A>

    my $IPaddrInDotNotation = TrafD::InetN2A($src_ip);

Converts number representation of IP address to convinient "dot" notation.

I.e. converts 2886729729 into string '172.16.0.1'

=back

=head1 SEE ALSO

L<trafd>, L<traflog>

=head1 HISTORY

A TrafD module appeared in trafd-accnt project http://sourceforge.net/projects/trafd-accnt
Written by Dmitry Kazarov 

=cut

our $errstr = '';

sub InetN2A($) {
    my $num = shift;
    $num = shift if ref($num);

    my $oct4 = $num & 0xFF;
    $num >>= 8;
    my $oct3 = $num & 0xFF;
    $num >>= 8;
    my $oct2 = $num & 0xFF;
    $num >>= 8;
    my $oct1 = $num & 0xFF;
    return ''.$oct1.'.'.$oct2.'.'.$oct3.'.'.$oct4;
}

my $checkTrafD3_4File = sub {
    local *File = shift @_;
    my ($FSize,$HdrSize,$RecSize) = @_;
    my $inData;

    die 'Can not seek file: '
	    unless seek File, 0, 0;

    while( read File,$inData,$HdrSize ) {
	my $NoRec = unpack "I",$inData;
	die 'Can not seek file: '
	    unless seek File, $NoRec*$RecSize, 1;
	my $curPos = tell File;
	return 1 if $curPos == $FSize;
	return 0 if $curPos >  $FSize;
    }
    return 0;
};


sub open {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $FileName = shift;

    unless( $FileName ) {
	$errstr = 'File name should be specified';
	return undef;
    }

    local *FH;

    unless( open FH, $FileName ) {
	$errstr = 'Can not open file '.$FileName.': '.$!;
	return undef;
    }

    my $fsize   = (stat FH)[7];
    my $version;
    if( $class eq 'TrafD' ) {
	eval {
	    $version =
		&$checkTrafD3_4File(*FH,$fsize,20,20)?3:
		&$checkTrafD3_4File(*FH,$fsize,20,24)?4:
		0;
	};
	$errstr = $@;
	return undef if $@;
	unless( $version ) {
	    $errstr = 'Wrong format of input file';
	    return undef;
	}
	$class = $version == 4 ? 'TrafD4' : 'TrafD3';
    }
    elsif( $class eq 'TrafD3' ) {
	$version = 3;
    }
    elsif( $class eq 'TrafD4' ) {
	$version = 4;
    }

    unless( seek FH, 0, 0 ) {
    	$errstr = 'Can not seek file "'.$FileName.'": '.$!;
	return undef;
    }

    my $self = {
	FileName => $FileName,
	FileHand => *FH,
	FileSize => $fsize,
	BPFTVers => $version,
    };

    return bless $self, $class;
}

# Reads Header of the Save.
#
# returns array
# ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec ) =
# Number or records
# Time of beginning of the statistics aquisition: seconds and microseconds
# time of end of the statistics aquisition: seconds and microseconds
# For 
# Common Header reader for TrafD4 and TrafD3
sub ReadHeader {
    my $this = shift;
    my $inData;
    local *File = $this->{FileHand};

    return () unless read File,$inData,20;
    return unpack "I5",$inData;
}

# Common "Save" reader for TrafD4 and TrafD3
sub ReadSave {
    my $this = shift;

    return ()
	unless
	    my ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec ) =
		$this->ReadHeader;

    my @Records;
    for( my $cntr=0;$cntr<$NoRec;$cntr++ ) {
	return () unless my @record = $this->ReadRecord;
	push @Records, [ @record ];
    }

    return ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec,@Records );
}

package TrafD4;
our @ISA = ('TrafD');

sub ReadRecord {
    my $this = shift;
    my $inData;

    local *File = $this->{FileHand};

    unless( read File,$inData,24 ) {
	$this->{errstr} = 'Can not read file: '.$!;
	return ();
    }

    my (
	$src_ip,$dst_ip,$proto,$ph0,
	$src_port,$dst_port,$ph1,$bytes,$dataBytes
    ) = unpack( "NNCCSSSLL", $inData );

    return ( $src_ip,$dst_ip,$proto,$src_port,$dst_port,$bytes,$dataBytes );
}

package TrafD3;
our @ISA = ('TrafD');

sub ReadRecord {
    my $this = shift;
    my $inData;

    local *File = $this->{FileHand};
    unless( read File,$inData,20 ) {
	$this->{errstr} = 'Can not read file: '.$!;
	return ();
    }

    my (
	$src_ip,$dst_ip,$proto,$who_srv,
	$port,$bytes,$dataBytes
    ) = unpack( "NNCCSLL", $inData );

    # Trafd3 is a bit funny about storing port numbers
    # 0 is equiv of traflog3 'client', -1 - of 'none'
    my $user = $port ? 0: -1;
    my $src_port = $who_srv & 1? $port : $user;
    my $dst_port = $who_srv & 2? $port : $user;

    return ( $src_ip,$dst_ip,$proto,$src_port,$dst_port,$bytes,$dataBytes );
}

#sub ReadBPFT4File(*) {
#    my $File = $_[0];
#    my ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec );
#    my $inData;
#    my @Records= ();
#    my ($SumDataBytes,$SumBytes) = (0,0);
#
#    return () unless read $File,$inData,20;
#
#    ( $NoRec,$STimeSec,$STimeUSec,$ETimeSec,$ETimeUSec ) =
#	unpack "I5",$inData;
#
#
#    for( my $rec_cntr=0;$rec_cntr<$NoRec;$rec_cntr++ ) {
#	die 'Can not read a record No'.$rec_cntr
#	    unless defined read $File,$inData,24;
#	my (
#	    $src_ip,$dst_ip,$proto,$ph0,
#	    $src_port,$dst_port,$ph1,$bytes,$dataBytes
#	) = unpack( "NNCCSSSLL", $inData );
#	push @Records, [
#	    $src_ip,$dst_ip,$proto,$src_port,$dst_port,$bytes,$dataBytes 
#	];
#	$SumBytes     += $bytes;
#	$SumDataBytes += $dataBytes;
#    }
#    return (
#	$SumDataBytes,$SumBytes,$NoRec,$STimeSec,
#	$STimeUSec,$ETimeSec,$ETimeUSec,
#	@Records
#    )
#}

1;
