/*
 *	Copyright (c) 1999 RISS-Telecom Networking Center
 *
 *	Copyright (c) 1993 The CAD lab of the
 *	Novosibirsk Institute of Broadcasting and Telecommunication
 *
 *	BPFT $Id: main.c,v 1.8 2004/02/25 19:10:08 stas_degteff Exp $
 *
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*	main.c - tcp/udp data traffic collector daemon	*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <paths.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#ifndef __linux__
#include <sys/syslimits.h>
#endif
#include <sys/stat.h>
#include <netinet/in.h>
#include <net/bpf.h>
#include <pcap.h>

#include "../include/interface.h"
#include "../include/pathnames.h"
#include "../include/traffic.h"
#include "../include/addrtoname.h"
#include "trafd.h"

#define USE_EXIT 0	/* Use exit() function or use return operator
			   in the 'main' function */

/*
 * Command line switches.
 * Left - default value, right - action, if switch has been used.
 */
int dflag = 0;	/* dump to compiled packet-matching code to stdout and stop */
int Oflag = 1;	/* don't run the packet-matching code optimizer if have bug */
int pflag = 0;	/* don't put the interface into promiscuous mode */
int rflag = 0;	/* attempt to resume data from safe file if exist */
int nflag = 0;
int Nflag = 0;
int fflag = 0;
int Xflag = 1;	/* collect 'extended' traffic info (hosts-protocol-ports) */

/* Global interrupts flags */
int flag_hup;	/* SIGHUP - drop collected data to tempfile */
int flag_int;	/* SIGINT - append collected data to savefile */
int flag_usr1;

/* Global variables */
char *program_name;	/* myself */
char *device_name = 0;	/* interface name */
char file_pid[PATH_MAX];	/* pid file */
char file_out[PATH_MAX];	/* dump file */
char file_backup[PATH_MAX];	/* save file */
char file_tmp[PATH_MAX];	/* temp file */
char file_fifo[PATH_MAX];	/* fifo file */
unsigned long minsize=MIN_SIZE_TO_SAVE;

/* Block of the interrupt drivers */
static void cleanup();
static void onhup();
static void onint();
static void onusr1();
static void onusr2();
static void onalarm();
static void onchld();

int is_symlink(const char*);
void check_stats();
char *version();

/* Length of saved portion of packet */
int snaplen = DEFAULT_SNAPLEN;

/*static*/ pcap_t *pd;

int
main(argc, argv)
	int argc;
	char **argv;
{
	struct bpf_program *parse();
	void bpf_dump(), usage(), printver(), check_stats();

	register int op, cnt = -1, i;
	struct bpf_program fcode;
	register char *cp, *infile = NULL, *cmdbuf, *DevFileName = NULL;
	FILE *fd;
	pcap_handler printer;
	u_char *pcap_userdata;
	char ebuf[PCAP_ERRBUF_SIZE];
	extern char *optarg;
	extern int optind, opterr;
	bpf_u_int32 localnet, netmask;

	if ((cp = strrchr(argv[0], '/')) != NULL)
			program_name = cp + 1;
	else
			program_name = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "c:df:F:i:OprVXm:")) != EOF)
		switch (op) {
		case 'c':
			cnt = atoi(optarg);
			break;
		case 'd':
			++dflag;
			break;
		case 'f':
			DevFileName = optarg;
			break;
		case 'F':
			infile = optarg;
			break;
		case 'i':
			device_name = optarg;
			break;
		case 'm':
			minsize = strtoul(optarg, NULL,10);
			break;
		case 'O':
			Oflag = 0;
			break;
		case 'p':
			++pflag;
			break;
		case 'r':
			++rflag;
			break;
		case 'X':
			Xflag=0;
			break;
		case 'V':
			printver();
			return 0;
			break;
		default:
			usage();
		}

	if (access(_PATH_VARRUN, R_OK|W_OK|X_OK) < 0) {  /* ITS4: ignore */
		snprintf(file_pid, PATH_MAX,"%s: access to %s", program_name,
			_PATH_VARRUN);
		perror(file_pid);
		exit(1);
	}
	if (access(PATH_TOSAVE, R_OK|W_OK|X_OK) < 0) {   /* ITS4: ignore */
		snprintf(file_backup, PATH_MAX, "%s: access to %s", program_name,
			PATH_TOSAVE);
		perror(file_backup);
		exit(1);
	}
	if (device_name == 0)
		if ((device_name = getenv("IFF_LISTEN")) == NULL)
			if ((device_name = pcap_lookupdev(ebuf)) == 0)
				error("%s", ebuf);
	if (DevFileName == NULL)
		DevFileName = device_name;

	/* Attach bpf interface to network interface */
	pd = pcap_open_live(device_name, snaplen, !pflag, 1000, ebuf);
	if (pd == NULL)
			error("%s", ebuf);

	i = pcap_snapshot(pd);
	if (snaplen < i) {
		warning("snaplen raised from %d to %d", snaplen, i);
		snaplen = i;
	}
	
	if (pcap_lookupnet(device_name, &localnet, &netmask, ebuf) < 0) {
		localnet = 0;
		netmask = 0;
		warning("%s", ebuf);
	}
							
	if (infile!=NULL)
		cmdbuf = read_infile(infile);
	else
		cmdbuf = copy_argv(&argv[optind]);

	if(pcap_compile(pd, &fcode, cmdbuf, Oflag, netmask) < 0)
			error("%s", pcap_geterr(pd));
	
	if (dflag) {
		bpf_dump(&fcode, dflag);
#if USE_EXIT
		exit(0);
#else
		return 0;
#endif
	}
	init_addrtoname(localnet, netmask);

	/* Catch signals */
	flag_int = flag_hup = flag_usr1 = 0;
	(void)signal(SIGTERM, cleanup);
	(void)signal(SIGQUIT, cleanup);
	(void)signal(SIGPIPE, SIG_IGN);
	(void)signal(SIGINT,  onint);
	(void)signal(SIGHUP,  onhup);
	(void)signal(SIGUSR1, onusr1);
	(void)signal(SIGUSR2, onusr2);
	(void)signal(SIGALRM, onalarm);
	(void)signal(SIGCHLD, onchld);

	/* Initialize file names */
	snprintf(file_pid, PATH_MAX, "%s%s.%s",  _PATH_VARRUN, program_name, device_name);
	snprintf(file_out, PATH_MAX, "%s%s.%s",  _PATH_VARTMP, program_name, DevFileName);
	snprintf(file_backup, PATH_MAX, "%s%s.%s", PATH_TOSAVE, program_name, DevFileName);
	snprintf(file_tmp, PATH_MAX, "%stmp.%s", PATH_TOSAVE, DevFileName);
	snprintf(file_fifo, PATH_MAX, "%s%s.%s", _PATH_TMP,    program_name, device_name);

	/* Check for symlinks */
	if( access(file_out,F_OK) ){
		/* create file & set access mask */
/*		fd = fopen(file_out,"w");
		fchmod(fd,TMP_UMASK);
		fclose(fd);
*/	}
	else if( is_symlink(file_out) )		/*ITS4: ignore*/
		error("%s is symlink", file_out);
	if( !access(file_backup,F_OK) && is_symlink(file_backup) )	/*ITS4: ignore*/
		error("%s is symlink", file_backup);
	if( !access(file_fifo,F_OK) && is_symlink(file_fifo) )		/*ITS4: ignore*/
		error("%s is symlink", file_fifo);

	/* Todo: Need check for old pidfile */
	if (access(file_pid, F_OK) == 0)  /* ITS4: ignore */
		error("unexpected pidfile, probable interface '%s' already listen",
		      device_name);

	check_stats();

	/* Jump to background */
	daemon(1, 0);
	if ((fd = fopen(file_pid, "w")) == NULL)	/* ITS4: ignore */
#if USE_EXIT
		exit(1);
#else
		return 1;
#endif

	fprintf(fd,"%d\n", getpid());	/* ITS4: ignore */
	fclose(fd);

/*	openlog(program_name, LOG_PID|LOG_CONS, LOG_DAEMON);*/
	openlog(program_name, LOG_PID|LOG_CONS, SYSLOG_FACILITY);	/*ITS4: ignore*/
	syslog(LOG_NOTICE, "(%s) traffic collector started", device_name);	/* ITS4: ignore */

	if (pcap_setfilter(pd, &fcode) < 0)
		error("%s", pcap_geterr(pd));
	
	printer = lookup_printer(pcap_datalink(pd));
	pcap_userdata = 0;
	/* Start read from bpf */
	if (!traf_init(rflag)) {
		if (pcap_loop(pd, cnt, printer, pcap_userdata))
			/* Clear exit */
			cleanup();
	} else
		errno = ENOMEM;

	pcap_close(pd);
	/* Abnormal termination */
	(void)syslog(LOG_ERR, "(%s) traffic collector aborted: %m",	/* ITS4: ignore */
		     device_name);
#if USE_EXIT
	exit(1);
#endif
	return 1;
}

/* make a clean exit on interrupts */
static void
cleanup()
{

	struct pcap_stat stat;

	traf_save(file_out, "w");
	if (!pcap_stats(pd, &stat))
		(void)syslog(LOG_INFO, "(%s) packets: received %d, dropped %d",	/* ITS4: ignore */
			     device_name, stat.ps_recv, stat.ps_drop);
	pcap_close(pd);
	unlink(file_pid);	/* ITS4: ignore */
	pd = NULL;
	(void)syslog(LOG_NOTICE, "(%s) traffic collector stopped", device_name); /* ITS4: ignore */
	exit(0);
}

#define	WAIT_ACTION	5

static void
onhup()
{
	alarm(WAIT_ACTION);
	flag_hup++;
}

static void
onint()
{
	alarm(WAIT_ACTION);
	flag_int++;
}

static void
onusr1()
{
	flag_usr1++;
}

static void
onusr2()
{
	flag_usr1 = 0;
	traf_pipe();
}

static void
onalarm()
{
	alarm(0);
	if (flag_hup) {
		flag_hup = 0;
		traf_save(file_out, "w");
	}
	if (flag_int) {
		flag_int = 0;
		traf_save(file_backup, "a");
		traf_clear();
	}
}

static void
onchld()
{
	if (wait(0) == -1)
		(void)syslog(LOG_WARNING, "(%s) wait: %m", device_name); /* ITS4: ignore */
}


void
printver()
{

	fprintf(stderr,	/* ITS4: ignore */
		"trafd v%s - tcp/udp data traffic collector daemon (BPFT project)\n\n",
		version() );
}

void
usage()
{
	printver();
	fprintf(stderr,	/* ITS4: ignore */
"Usage: %s [-dOprVX] [-c count] [-i iface] [-f ext] [-m minsize] [-F file | expr]\n", program_name);
	exit(-1);
}


/* Check status of the interface and print warnings if needed */
void check_stats()
{	struct pcap_stat pcap_status={0,0,0};

	if( pcap_stats(pd, &pcap_status) )
		warning("(%s) packets dropping checks not avaiable (pcap_stats(): %s)",
			     device_name, pcap_geterr(pd) );
	else if( pcap_status.ps_ifdrop )
		warning("(%d) drops by interface %s not yet supported, data loss is possible",
			     pcap_status.ps_ifdrop, device_name );
}

int is_symlink(path)
	const char *path;
{
	struct stat st,lst;

	lstat( path, &lst );	/*ITS4: ignore*/
	stat( path, &st );	/*ITS4: ignore*/
	return ( lst.st_dev!=st.st_dev || lst.st_ino!=st.st_ino );
}
